﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using PalmSens;
using PalmSens.Core.Simplified.InternalStorage;
using PalmSens.Data;
using PalmSens.Devices;

namespace PSSDKInternalStorageExample
{
    /// <summary>
    ///     Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        /// <summary>
        ///     The contents of the current directory on the internal storage
        /// </summary>
        private readonly IDictionary<int, IInternalStorageItem> _contents;

        /// <summary>
        ///     The connected PalmSens & EmStat devices
        /// </summary>
        private Device[] _connectedDevices = new Device[0];

        /// <summary>
        ///     The current root item.
        /// </summary>
        private IInternalStorageFolder _root;

        public MainWindow()
        {
            InitializeComponent();
            DiscoverConnectedDevices();
            _contents = new Dictionary<int, IInternalStorageItem>();
        }

        private void btnConnect_Click(object sender, RoutedEventArgs e)
        {
            if (!psCommSimpleWPF.Connected) //Determine whether a device is currently connected
            {
                if (cmbDevices.SelectedIndex == -1)
                    return;

                try
                {
                    //Connect to the device selected in the devices combobox control
                    psCommSimpleWPF.Connect(_connectedDevices[cmbDevices.SelectedIndex]);
                    LogMessage($"Connected to {psCommSimpleWPF.ConnectedDevice.ToString()}");
                }
                catch (Exception ex)
                {
                    LogMessage(ex.Message);
                }
            }
            else
            {
                psCommSimpleWPF.Disconnect(); //Disconnect from the connected device
                LogMessage("Disconnected");
            }

            //Update UI based on connection status
            cmbDevices.IsEnabled = !psCommSimpleWPF.Connected;
            btnRefresh.IsEnabled = !psCommSimpleWPF.Connected;
            btnConnect.Content = psCommSimpleWPF.Connected ? "Disconnect" : "Connect";
            btnListFiles.IsEnabled = psCommSimpleWPF.Connected && psCommSimpleWPF.Capabilities.SupportsStorage;

            if (psCommSimpleWPF.Connected && psCommSimpleWPF.Capabilities.SupportsStorage)
                GetSpaceUsed();
            else if (psCommSimpleWPF.Connected)
                LogMessage("The connected device does not support internal storage features");
        }

        private void btnListFiles_Click(object sender, RoutedEventArgs e)
        {
            // Clear all nodes
            LogMessage("Reading contents from: Root");

            _root = psCommSimpleWPF.GetInternalStorageBrowser().GetRoot();

            LoadChildren(_root);
        }

        /// <summary>
        ///     Handles the Click event of the btnOpen control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="EventArgs" /> instance containing the event data.</param>
        private void btnOpen_Click(object sender, RoutedEventArgs e)
        {
            OpenItem();
        }

        private void btnRefresh_Click(object sender, RoutedEventArgs e)
        {
            DiscoverConnectedDevices(); ////Add connected devices to cmbDevices combo box
        }

        /// <summary>
        ///     Discovers the connected PalmSens & EmStat devices and adds them to the combobox control.
        /// </summary>
        private void DiscoverConnectedDevices()
        {
            cmbDevices.Items.Clear();

            _connectedDevices = psCommSimpleWPF.ConnectedDevices; //Discover connected devices

            foreach (var d in _connectedDevices)
                cmbDevices.Items.Add(d.ToString()); //Add connected devices to control

            var nDevices = cmbDevices.Items.Count;
            cmbDevices.SelectedIndex = nDevices > 0 ? 0 : -1;
            LogMessage($"Found {nDevices} device(s).");

            btnConnect.IsEnabled = nDevices > 0;
        }

        /// <summary>
        ///     Gets the % of space used on the device.
        /// </summary>
        /// <exception cref="System.NotImplementedException"></exception>
        private void GetSpaceUsed()
        {
            var total = (int) psCommSimpleWPF.Comm.ClientConnection.GetDeviceSize();
            var free = (int) psCommSimpleWPF.Comm.ClientConnection.GetDeviceFree();

            var perc = Math.Min((total - free) / (double) total, 1.0);
            LogMessage($"{(int) Math.Round(perc * 100.0):F2}% of storage space used.");
        }

        private void LbInternalStorage_OnSelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            if (lbInternalStorage.SelectedIndex == -1) return;

            var index = lbInternalStorage.SelectedIndex;
            if (_contents.Count <= 1) index++;

            numUDTargetItem.Text = index.ToString();
        }

        /// <summary>
        ///     List the internal storage item
        /// </summary>
        /// <param name="item"></param>
        /// <param name="index"></param>
        private void ListInternalStorageItem(IInternalStorageItem item, int index)
        {
            var type = item.ItemType == DeviceFileType.Folder ? "<Dir>" : "<File>";
            lbInternalStorage.Items.Add($"{index}. {type} {item.Name}");
        }

        /// <summary>
        ///     List the parent item
        /// </summary>
        /// <param name="parent"></param>
        private void ListPreviousItem(IInternalStorageItem parent)
        {
            _contents[0] = parent;
            lbInternalStorage.Items.Add($"0. <Back> {parent.Name}");
        }

        /// <summary>
        ///     Add the selected item children to the form
        /// </summary>
        /// <param name="item"></param>
        private void LoadChildren(IInternalStorageItem item)
        {
            if (item.ItemType == DeviceFileType.Measurement) return;

            lbInternalStorage.Items.Clear();
            _contents.Clear();

            var folder = (IInternalStorageFolder) item;
            var files = folder.GetFiles();
            var subFolders = folder.GetSubFolders();
            var totalItems = files.Count + subFolders.Count;
            var hasParent = !ReferenceEquals(item, _root);

            numUDTargetItem.IsEnabled = totalItems > 0;
            btnOpen.IsEnabled = numUDTargetItem.IsEnabled;
            numUDTargetItem.Text = totalItems > 0 ? "1" : "0";

            LogMessage($"Loading {totalItems} storage items.");

            if (hasParent)
                ListPreviousItem(folder.Parent);
            var currentIndex = LoadChildrenItems(subFolders, 1);
            LoadChildrenItems(files, currentIndex);
        }

        /// <summary>
        ///     Load the child items and add to form.
        /// </summary>
        /// <param name="items"></param>
        /// <param name="index"></param>
        /// <returns></returns>
        private int LoadChildrenItems(IEnumerable<IInternalStorageItem> items, int index)
        {
            foreach (var item in items.OrderBy(i => i.Name))
            {
                ListInternalStorageItem(item, index);
                _contents[index] = item;
                index++;
            }

            return index;
        }

        /// <summary>
        ///     Log an error message
        /// </summary>
        /// <param name="exception"></param>
        private void LogMessage(Exception exception)
        {
            LogMessage($"An error has occurred: {exception.Message}");
        }

        private void LogMessage(string message)
        {
            lbConsole.SelectedIndex = lbConsole.Items.Add(message);
            lbConsole.ScrollIntoView(lbConsole.SelectedItem);
        }

        private void numUDTargetItem_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.Key >= Key.D0 && e.Key <= Key.D9 || e.Key >= Key.NumPad0 && e.Key <= Key.NumPad9) return;

            e.Handled = true;
        }

        /// <summary>
        ///     Open the item
        /// </summary>
        private void OpenItem()
        {
            try
            {
                var index = int.Parse(numUDTargetItem.Text);

                if (index < 0 || index > _contents.Count)
                {
                    LogMessage("The item at the specified position does not exist");
                    return;
                }

                var item = _contents[index];

                if (item.ItemType == DeviceFileType.Folder)
                {
                    LogMessage($"Opening path: {item.Name}");
                    LoadChildren(item);
                }
                else
                {
                    var file = (IInternalStorageFile) item;

                    var m = file.GetMeasurement(MeasType.Overlay);

                    LogMessage(m != null ? $"Load file '{item.Name}', with method '{m.Method}'" : $"Could not load file '{item.Name}'");
                }
            }
            catch (Exception exception)
            {
                LogMessage(exception);
            }
        }

        private void TextBlock_MouseDown(object sender, MouseButtonEventArgs e)
        {
            if (e.ClickCount < 2) return;

            OpenItem();
        }
    }
}